/* Panorama_Tools	-	Generate, Edit and Convert Panoramic Images
   Copyright (C) 1998,1999,2000 - Helmut Dersch  der@fh-furtwangen.de
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

/*------------------------------------------------------------*/

#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>
#include <X11/keysym.h>
#include <X11/Xatom.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdint.h>

#include "filter.h"
#include "panolib.h"
#include "utils_image.h"

#define DEF_VIEW_WIDTH  400
#define DEF_VIEW_HEIGHT 300

char text[] = "PTViewer";

Display 	*disp_g;
Window 		window_g;
int 		screen_g;
GC 		gc;
Visual    	*visual_g;
unsigned int  	depth_g;


static int copy_rgb_to_zpixmap (XImage *dest, const ui_image_t *src)
{
	uint32_t *src_data;

	uint32_t src_r;
	uint32_t src_g;
	uint32_t src_b;

	uint32_t dest_r;
	uint32_t dest_g;
	uint32_t dest_b;
	uint32_t dest_pixel_value;

	uint32_t x;
	uint32_t y;
	uint32_t pixel;

	uint32_t dest_r_maxval;
	uint32_t dest_g_maxval;
	uint32_t dest_b_maxval;

	uint32_t dest_r_offset;
	uint32_t dest_g_offset;
	uint32_t dest_b_offset;

	dest_r_offset = 0;
	dest_g_offset = 0;
	dest_b_offset = 0;
	dest_r_maxval = dest->red_mask;
	dest_g_maxval = dest->green_mask;
	dest_b_maxval = dest->blue_mask;
	for (x = 0; x < dest->depth; x++)
	{
		if ((dest_r_maxval & 0x01) == 0)
		{
			dest_r_offset++;
			dest_r_maxval >>= 1;
		}
		if ((dest_g_maxval & 0x01) == 0)
		{
			dest_g_offset++;
			dest_g_maxval >>= 1;
		}
		if ((dest_b_maxval & 0x01) == 0)
		{
			dest_b_offset++;
			dest_b_maxval >>= 1;
		}
	}

	src_data = (uint32_t *) *src->data;

	pixel = 0;
	for (y = 0; y < dest->height; y++)
	{
		for (x = 0; x < dest->width; x++, pixel++)
		{
			int32_t bytenum;

			src_r = src->data[0][pixel];
			src_g = src->data[1][pixel];
			src_b = src->data[2][pixel];

			dest_r = dest_r_maxval * src_r / 255;
			dest_g = dest_g_maxval * src_g / 255;
			dest_b = dest_b_maxval * src_b / 255;

			dest_pixel_value = 0
				| ((dest_r << dest_r_offset) & dest->red_mask)
				| ((dest_g << dest_g_offset) & dest->green_mask)
				| ((dest_b << dest_b_offset) & dest->blue_mask);

			for (bytenum = 0; bytenum < (dest->depth / 8); bytenum++)
			{
				dest->data[(pixel * dest->bits_per_pixel / 8) + bytenum] = 
					(dest_pixel_value >> (dest->bits_per_pixel - (8 * (bytenum + 1)))) & 0xFF;
			}
		}
	}

	return (0);
} /* int copy_rgb_to_zpixmap */

static int draw_window (const ui_image_t *img)
{
  static XImage *ximage = NULL;

  if ((ximage != NULL) && ((img->width != ximage->width) || (img->height != ximage->height)))
  {
    free (ximage->data);
    ximage->data = NULL;
    XDestroyImage (ximage);
    ximage = NULL;
  }

  if (ximage == NULL)
  {
    ximage = XCreateImage(disp_g, visual_g, depth_g, ZPixmap, 0,
	NULL, img->width, img->height,
	8, 0);
    if (ximage == NULL)
      return (-1);

    ximage->byte_order= MSBFirst;

    ximage->data = (char *) malloc (img->width * img->height * depth_g / 8);
    if (ximage->data == NULL)
    {
      XDestroyImage (ximage);
      ximage = NULL;
      return (-1);
    }
  }

  copy_rgb_to_zpixmap (ximage, img);
  XPutImage(disp_g, window_g, gc, ximage, 0, 0,
      0, 0, ximage->width, ximage->height);

  return (0);
} /* int draw_window */

#define ZOOM_SPEED 1.03

static void zoom_in (double *fov)
{
  *fov = *fov / ZOOM_SPEED;
  if (*fov < 10.5)
    *fov = 10.5;
} /* void zoom_in */

static void zoom_out (double *fov)
{
  *fov = *fov * ZOOM_SPEED;
  if (*fov > 165.0)
    *fov = 165.0;
} /* void zoom_out */

int main( int argc, char** argv )
{
  unsigned long color_fg, color_bg;
  int done;
  static Atom           	proto_atom= None, delete_atom= None;

  long event_mask;

  double fov = 70.0;
  double yaw = 0.0;
  double pitch = 0.0;

  /* Detect mouse movement direction */
  unsigned int btn_current_pressed = 0;
  int btn_reference_pos_x = 0;
  int btn_reference_pos_y = 0;
  int btn_current_pos_x = 0;
  int btn_current_pos_y = 0;

  ui_image_t *pano;
  ui_image_t *view;


  // Set up display
  disp_g   = XOpenDisplay("");
  screen_g = DefaultScreen(disp_g);
  depth_g  = DefaultDepth(disp_g, screen_g);

  color_bg = WhitePixel(disp_g,screen_g);
  color_fg = BlackPixel(disp_g,screen_g);

  // Load panoramic image  	
  if( argc < 2 )
  {
    PrintError("No image file supplied");
    exit(0);
  }

  if( strcmp( argv[1], "-h" ) == 0 ) /* FIXME: Use getopt */
  {
    PrintError( "%s\n %s\nUsage: PTViewer Imagefile", "PTViewer", LONGVERSION );
    exit(0);
  }

  pano = ui_create_file (argv[1]);
  if (pano == NULL)
  {
    PrintError ("Unable to read pano file.");
    exit (0);
  }

  printf ("Panorama `%s', %ux%u pixels, loaded.\n",
      argv[1], pano->width, pano->height);

  view = ui_create (DEF_VIEW_WIDTH, DEF_VIEW_HEIGHT); /* FIXME: Possibly chose another size */
  if (view == NULL)
  {
    PrintError ("Unable to create view.");
    exit (0);
  }

  window_g = XCreateSimpleWindow(disp_g,
      DefaultRootWindow(disp_g),
      0, 0, /* Position */
      view->width, view->height,
      5,
      color_fg, color_bg);

  {
    XSizeHints hint;

    memset (&hint, '\0', sizeof (hint));
    hint.width  = view->width;
    hint.height = view->height;
    hint.flags = PSize;

    XSetStandardProperties(disp_g,
	window_g,
	text,
	text,
	None,
	argv,
	argc,
	&hint);
  }

  gc = XCreateGC(disp_g,window_g,0,0);
  XSetBackground(disp_g,gc,color_bg);
  XSetForeground(disp_g,gc,color_fg);

  event_mask = ButtonPressMask | ButtonReleaseMask | ButtonMotionMask
    | KeyPressMask | KeyReleaseMask | StructureNotifyMask
    | EnterWindowMask | LeaveWindowMask| ExposureMask;

  XSelectInput(disp_g, window_g, event_mask);
  XMapRaised(disp_g,window_g);

  proto_atom 	= XInternAtom(disp_g, "WM_PROTOCOLS", False);
  delete_atom = XInternAtom(disp_g, "WM_DELETE_WINDOW", False);
  if ((proto_atom != None) && (delete_atom != None))
    XChangeProperty(disp_g, window_g, proto_atom, XA_ATOM, 32,
	PropModeReplace, (unsigned char *)&delete_atom,
	1);
  visual_g  = DefaultVisual(disp_g, screen_g);

  pl_extract_view (view, pano, pitch, yaw, fov, BILINEAR);
  draw_window (view);

  done = 0;
  while (done == 0)
  {
    int isChanged = 0;
    XEvent event;

    memset (&event, '\0', sizeof (event));

    if (btn_current_pressed == 0)
    {
      /* No button is pressed -> block */
      XMaskEvent (disp_g, event_mask, &event);
    }
    else
    {
      /* A button is pressed -> do not block */
      XCheckMaskEvent (disp_g, event_mask, &event);
    }

    switch (event.type)
    {
      case 0: /* XCheckMaskEvent didn't return an event */
	break;

      case Expose:
	{
	  XExposeEvent *eev = (XExposeEvent *) &event;

	  while (eev->count != 0)
	  {
	    if (!XCheckTypedEvent (disp_g, Expose, &event))
	    {
	      eev = NULL;
	      break;
	    }
	    /* do nothing */
	  }

	  if (eev != NULL)
	    isChanged = 1;
	}
	break;

      case EnterNotify:
	break;

      case LeaveNotify:
	break;

      case ConfigureNotify:
	{
	  XConfigureEvent *cev = (XConfigureEvent *) &event;

	  ui_destroy (view);
	  view = ui_create (cev->width, cev->height);
	  isChanged = 1;
	}
	break;

      case KeyPress:
	{
	  XKeyEvent *kev = (XKeyEvent *) &event;
	  char buffer[64];
	  int buffer_len = sizeof (buffer);
	  KeySym ks;

	  buffer_len = XLookupString(kev, buffer, sizeof (buffer), &ks, NULL);

	  if (buffer_len >= sizeof (buffer))
	    buffer_len = sizeof (buffer) - 1;
	  buffer[buffer_len] = '\0';

	  switch (ks)
	  {
	    case XK_Shift_L:
	    case XK_Shift_R:
	      zoom_in (&fov);
	      isChanged = 1;
	      break;

	    case XK_Control_L:
	    case XK_Control_R:
	      zoom_out (&fov);
	      isChanged = 1;
	      break;

	    case XK_Up:
	      pitch += 5.0;
	      if (pitch > 90.0)
		pitch = 90.0;
	      isChanged = 1;
	      break;

	    case XK_Down:
	      pitch -= 5.0;
	      if (pitch < -90.0)
		pitch = -90.0;
	      isChanged = 1;
	      break;

	    case XK_Right:
	      yaw += 5.0;
	      if (yaw > 180.0)
		yaw -= 360.0;
	      isChanged = 1;
	      break;

	    case XK_Left:
	      yaw -= 5.0;
	      if (yaw <= -180.0)
		yaw += 360.0;
	      isChanged = 1;
	      break;

	    case XK_q:
	      done++;
	      break;

	    default:
	      printf ("Unhandeled KeySym 0x%02llx, buffer = %s\n", (uint64_t) ks, buffer);
	  }
	}
	break;

#if 0
      case KeyRelease:
	printf ("Ignoring KeyRelease event.\n");
	break;
#endif

      case ButtonPress:
      /* case ButtonRelease: */
	{
	  XButtonEvent *bev = (XButtonEvent *) &event;

	  btn_current_pressed = bev->button;
	  btn_current_pos_x = btn_reference_pos_x = bev->x;
	  btn_current_pos_y = btn_reference_pos_y = bev->y;
	}
	break;

      case ButtonRelease:
	btn_current_pressed = 0;
	/* Re-draw the image (when no button is pressed better interpolation is
	 * used. */
	isChanged = 1;
	break;

      case MotionNotify:
	{
	  XMotionEvent *mev = (XMotionEvent *) &event;

	  do
	  {
	    btn_current_pos_x = mev->x;
	    btn_current_pos_y = mev->y;
	  }
	  while (XCheckTypedEvent (disp_g, MotionNotify, &event));
	}
	break;

      default:
	printf ("Unhandled event type: 0x%02x\n",
	    event.type);
    } /* switch (event.type) */

    /*
     * Check for pressed buttons and change view accordingly
     */
    if (btn_current_pressed == Button1)
    {
      double yaw_diff = (btn_current_pos_x - btn_reference_pos_x) / 24.0;
      if (yaw_diff > 90.0)
	yaw_diff = 90.0;
      else if (yaw_diff < -90.0)
	yaw_diff = -90.0;

      yaw += yaw_diff;
      if (yaw > 180.0)
	yaw -= 360.0;
      else if (yaw <= -180.0)
	yaw += 360.0;

      pitch -= (btn_current_pos_y - btn_reference_pos_y) / 24.0;
      if (pitch > 90.0)
	pitch = 90.0;
      else if (pitch < -90.0)
	pitch = -90.0;

      isChanged = 1;
    }
    else if (btn_current_pressed == Button3)
    {
      double zoom_diff = (btn_current_pos_y - btn_reference_pos_y) / 64.0;
      double zoom_ratio = pow (ZOOM_SPEED, zoom_diff);

      fov *= zoom_ratio;
      if (fov < 10.5)
	fov = 10.5;
      else if (fov > 165.0)
	fov = 165.0;

      isChanged = 1;
    }

    if (isChanged != 0)
    {
      pl_extract_view (view, pano, pitch, yaw, fov,
	  (btn_current_pressed == 0) ? BILINEAR : NNEIGHBOUR);
      draw_window (view);
    }
  } /* while (done == 0) */

  XFreeGC(disp_g,gc);
  XDestroyWindow(disp_g, window_g);
  XCloseDisplay(disp_g);

  return (0);
} /* int main */
	





// Error reporting

void  PrintError( char* fmt, ...)
{
	va_list	ap;
	char message[257];
	
	va_start(ap, fmt);
	vsprintf(message, fmt, ap);
	va_end(ap);
	
	printf("%s\n", message);
}

void**  mymalloc( long numBytes )					// Memory allocation, use Handles
{
	char **mem;
	
	mem = (char**)malloc( sizeof(char*) );			// Allocate memory for pointer
	if(mem == NULL)
		return (void**)NULL;
	else
	{
		(*mem) = (char*) malloc( numBytes );		// Allocate numBytes
		if( *mem == NULL )
		{
			free( mem );
			return (void**)NULL;
		}
		else
			return (void**)mem;
	}
}

void 	myfree( void** Hdl )						// free Memory, use Handles
{
	free( (char*) *Hdl );
	free( (char**) Hdl );
}		


void SetImageDefaults(Image *im)
{
	im->data 			= NULL;
	im->bytesPerLine	= 0;
	im->width			= 0;
	im->height			= 0;
	im->dataSize		= 0;
	im->bitsPerPixel	= 0;
	im->format			= 0;
	im->dataformat		= _RGB;
	im->hfov			= 0.0;
	im->yaw				= 0.0;
	im->pitch			= 0.0;
	im->roll			= 0.0;
//	SetCorrectDefaults( &(im->cP) );
	*(im->name)			= 0;
}

// expand image from 3 to 4 bits per pixel. No pad bytes allowed.
// Memory must be allocated
void ThreeToFourBPP( Image *im )
{
	register int x,y,c1,c2;

	if( im->bitsPerPixel == 32 || im->bitsPerPixel == 64) // Nothing to do
		return;
	
	
	
	if( im->bitsPerPixel == 24 )	// Convert to 4byte / pixel
	{
		for( y = im->height-1; y>=0; y--)
		{
			for( x= im->width-1; x>=0; x--)
			{
				c1 = (y * im->width + x) * 4;
				c2 = y * im->bytesPerLine + x * 3;
				(*(im->data))[c1++] = UCHAR_MAX;
				(*(im->data))[c1++] = (*(im->data))[c2++];
				(*(im->data))[c1++] = (*(im->data))[c2++];
				(*(im->data))[c1++] = (*(im->data))[c2++];
			}
		}
		im->bitsPerPixel = 32;
		im->bytesPerLine = im->width * 4;
	}
	else if( im->bitsPerPixel == 48 ) // Convert to 8byte / pixel
	{
		for( y = im->height-1; y>=0; y--)
		{
			for( x= im->width-1; x>=0; x--)
			{
				c1 = (y * im->width + x) * 4;
				c2 = y * im->bytesPerLine/2 + x * 3;
				((USHORT*)(*(im->data)))[c1++] = USHRT_MAX;
				((USHORT*)(*(im->data)))[c1++] = ((USHORT*)(*(im->data)))[c2++];
				((USHORT*)(*(im->data)))[c1++] = ((USHORT*)(*(im->data)))[c2++];
				((USHORT*)(*(im->data)))[c1++] = ((USHORT*)(*(im->data)))[c2++];
			}
		}
		im->bitsPerPixel = 64;
		im->bytesPerLine = im->width * 8;
	}
	im->dataSize = im->height * im->bytesPerLine;
}

/*
 * vim: set shiftwidth=2 tabstop=8 softtabstop=2 :
 */
